import yodu
from examples.steem.helpers import load_test_items, load_test_actions
from models.request import Request

recommender = yodu.create_recommender(name="example")

recommender = yodu.get_recommender(name="example")

# Add Items to Recommender
items = load_test_items()
recommender.item.add(items)

# Add Actions to Recommender
actions = load_test_actions()
recommender.action.add(actions)

# Enable Yodu's built-in Providers
recommender.provider.add(name="top_item_by_user_action")
# Add provider from Source (Coming Soon)

algo_spec = {
    "providers": {
        "TOP_BY_PREVIOUS_LIKED_SOURCES": {
            "provider": "top_item_by_user_action",
            "description": "Get top items from past top liked sources",
            "duration": "24h",
            "config": {"action_type": "LIKE", "tag": "source"},
            "weight": 1,
        },
        "TOP_BY_PREVIOUS_LIKED_CATEGORIES": {
            "provider": "top_item_by_user_action",
            "description": "Get top items from past top liked categories",
            "duration": "30h",
            "config": {"action_type": "LIKE", "tag": "category"},
            "weight": 1,
        },
        "TOP_BY_PREVIOUS_READ_CATEGORIES": {
            "provider": "top_item_by_user_action",
            "description": "Get top items from past top READ categories",
            "duration": "30h",
            "config": {"action_type": "READ", "tag": "category"},
            "weight": 1,
        },
    },
    "filters": {
        "PAST_ACTION": {
            "provider": "get_past_user_item_action",
            "description": "Filter items if a user has performed any action on the item",
            "duration": "24h",
            "config": {"action_type": "ALL"},
        },
    },
}
recommender.algo_spec.set(algo_spec=algo_spec)

args = {"days_ago": "7"}
request = Request(user_id="test_user_1", args=args)

items = recommender.get_items(request=request)



import datetime
import os
import shutil

import dateparser


def time_ago_to_date(time_ago_str):
    if time_ago_str == "now":
        date = datetime.datetime.now()
    else:
        date = dateparser.parse(time_ago_str)
    date = date.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    return date


def move_dir(base_dir, source_dir, destination_dir):
    # fetch all files
    source_dir = base_dir + "/" + source_dir
    destination_dir = base_dir + "/" + destination_dir
    os.makedirs(os.path.dirname(destination_dir), exist_ok=True)
    try:
        # Delete provider if exists
        shutil.rmtree(destination_dir)
    except:
        pass
    shutil.copytree(source_dir, destination_dir)



from queue import Queue

import ray

from yodu import ESClient
from yodu.algo_spec.helper import AlgoSpecHelper
from yodu.helpers.action_helper import ActionHelper
from yodu.helpers.item_helper import ItemHelper
from yodu.models.request import Request
from yodu.provider.helper import ProviderHelper

NUM_THREADS = 5
que = Queue()
threads_list = list()


class Recommender:
    __recommender_name = None
    __providers = None
    __indices = None
    __es_client = None

    item = None
    action = None
    provider = None
    algo_spec = None

    def __init__(self, name, indices, algo_spec=None, es_client=None):
        self.recommender_name = name
        self.__indices = indices
        self.__es_client = es_client
        if self.__es_client is None:
            self.__es_client = ESClient().get_client()
        self.provider = ProviderHelper(recommender_name=self.recommender_name)
        self.item = ItemHelper(
            index_name=self.__indices["items"], es_client=self.__es_client
        )
        self.action = ActionHelper(
            index_name=self.__indices["actions"], es_client=self.__es_client
        )
        self.algo_spec = AlgoSpecHelper(recommender_name=self.recommender_name)

    def get_docs(self, item_index, item_ids: list):
        hits = self.__es_client.mget(index=item_index, body={"ids": item_ids})
        items = []
        for doc in hits.body["docs"]:
            items.append(doc["_source"])
        return items

    def get_indices(self):
        return self.__indices

    def get_items(self, request: Request):
        """
        Algo:
        Add all providers to threads.
        Wait for x seconds
        return results
        """
        algo_spec = self.algo_spec.load()

        result_ids = []
        for provider_name, provider_dict in algo_spec["providers"].items():
            args = provider_dict["config"]
            provider_obj = self.provider.load_provider(
                name=provider_dict["provider"]
            ).remote(indices=self.__indices, name=provider_name)
            args["user_id"] = request.user_id
            result_ids.append(
                provider_obj.get_items.remote(config=args, request=request)
            )

        results = ray.get(result_ids)
        top_items_ids = {}
        top_items_providers = {}
        for provider_results in results:
            provider_name = provider_results[0]
            for item_id in provider_results[1]:
                if item_id in top_items_ids:
                    top_items_ids[item_id] = top_items_ids[item_id] + 1
                    top_items_providers[item_id].append(provider_name)
                else:
                    top_items_ids[item_id] = 1
                    top_items_providers[item_id] = [provider_name]
        top_items_ids = dict(
            sorted(
                top_items_ids.items(), key=lambda item: item[1], reverse=True
            )
        )
        if len(top_items_ids) > 0:
            top_items_ids = list(top_items_ids.keys())[: request.limit]
            items = self.get_docs(
                item_index=self.__indices["items"],
                item_ids=top_items_ids,
            )
            for item in items:
                item["source_provider"] = top_items_providers[item["id"]][0]
            return items
        return None



import json
from typing import Dict, Any

import ray

from yodu import ESClient
from yodu.models.request import Request
from yodu.provider.provider_base import ProviderBase
from yodu.utils.utils import time_ago_to_date

meta = {
    "name": "TopItemsByUserAction",
    "version": "0.0.1",
    "author": "Shashank Agarwal",
    "description": "Return top items for a source by action_type \
        Example: Return top liked items for a given source \
        Example: Return top read items for a given source \
        Example: Return top read items for a given category",
    "props": {
        "action_type": {"type": "str", "source": "request.props.action_type"},
        "days_ago": {
            "type": "int",
            "source": "request.props.days_ago",
            "default": 1,
        },
        "user_id": {"type": "str", "source": "request.props.user_id"},
        "tag": {"type": "str", "source": "request.props.tag"},
        "next_token": {"type": "str", "source": "request.next_token"},
        "limit": {"type": "int", "source": "request.limit", "default": 10},
        "offset": {"type": "int", "source": "request.offset", "default": 0},
    },
}


@ray.remote
class Provider(ProviderBase):
    __es_client = None
    __indices = None
    __name = None
    __meta = None

    def __init__(self, indices=Dict, name=str, **data: Any):
        super().__init__(meta=meta, data=data)
        self.__name = name
        if not self.__es_client:
            self.__es_client = ESClient()
        self.__indices = indices
        self.__meta = meta

    def get_items(self, config: Dict, request: Request):
        """
        :param config:
        :param request:
        :return:
        """
        query = """
        {
          "query": {
                "bool": {
                    "must": {
                        "term": {
                            "type.keyword": "ACTION_TYPE_PLACEHOLDER"
                        }
                    },
                    "filter": {
                        "range": {
                            "created_at": {
                                "gt": "DATE_AGO_PLACEHOLDER",
                                "lt": "DATE_NOW_PLACEHOLDER"
                            }
                        }
                    }
                }
            },
          "aggs": {
            "top_tag_by_action": {
              "terms": {
                "field": "props.TAG_PLACEHOLDER.keyword",
                "size": LIMIT_PLACEHOLDER
              },
              "aggs": {
                "top_action_hits": {
                  "top_hits": {
                    "sort": [
                      {
                        "created_at": {
                          "order": "desc"
                        }
                      }
                    ],
                    "_source": {
                      "includes": [ "item_id", "user_id","type" , "value", "props" ]
                    },
                    "size": ITEMS_PER_ACTION
                  }
                }
              }
            }
          }
        }
        """
        input_dict = self.build_input(config=config, request=request)
        time_now = time_ago_to_date("now")
        time_ago = time_ago_to_date(str(input_dict["days_ago"]) + " days ago")
        query = query.replace("DATE_NOW_PLACEHOLDER", time_now)
        query = query.replace("DATE_AGO_PLACEHOLDER", time_ago)
        query = query.replace(
            "ACTION_TYPE_PLACEHOLDER", input_dict["action_type"]
        )
        query = query.replace("TAG_PLACEHOLDER", input_dict["tag"])
        query = query.replace("LIMIT_PLACEHOLDER", str(input_dict["limit"]))
        query = query.replace("ITEMS_PER_ACTION", str(5))
        res = json.loads(query)
        results = self.execute(
            es_client=self.__es_client, indices=self.__indices, query=res
        )
        return [self.__name, results]



from typing import Dict

from yodu import ESClient
from yodu.models.request import Request


class ProviderBase:
    __meta = None
    __data = None

    def __init__(self, meta, data):
        self.__meta = meta
        self.__data = data

    def get_items(self, **kwprops):
        pass

    def get_meta(self):
        return self.__meta

    def build_input(self, config: Dict, request: Request):
        input_dict = {}
        """
        Get all values from request
        Then get all values from config
        Rest all values from meta
        """
        if request.limit:
            input_dict["limit"] = request.limit
        if request.offset:
            input_dict["offset"] = request.offset
        if request.user_id:
            input_dict["user_id"] = request.user_id
        if request.props:
            for key, val in request.props.items():
                input_dict[key] = val
        for name, val in config.items():
            if name not in input_dict:
                input_dict[name] = config[name]
        for name, val in self.__meta["props"].items():
            if (
                name not in input_dict
                and name in self.__meta["props"]
                and "default" in self.__meta["props"][name]
            ):
                input_dict[name] = self.__meta["props"][name]["default"]
        return input_dict

    def execute(self, es_client: ESClient, indices, query):
        action_index = indices["actions"]
        hits = es_client.get_client().search(
            index=action_index, size=0, **query
        )
        res = hits.body["aggregations"]["top_tag_by_action"]["buckets"]
        """
        returns a list of top categories.
        We then get items from each of these categories and rank by count
        """
        items_dict = {}
        for hit in res:
            for row in hit["top_action_hits"]["hits"]["hits"]:
                item = row["_source"]
                if "item_id" in item:
                    if item["item_id"] in items_dict:
                        items_dict[item["item_id"]] = (
                            items_dict[item["item_id"]] + 1
                        )
                    else:
                        items_dict[item["item_id"]] = 1
        items_dict = dict(
            sorted(items_dict.items(), key=lambda item: item[1], reverse=True)
        )
        top_items = list(items_dict.keys())
        return top_items



from datetime import datetime
from typing import Optional, Dict
from uuid import uuid4

from pydantic import BaseModel, Field


class Action(BaseModel):
    # Unique ID for the item. If no ID is provided, a uuid4 is generated.
    id: str = Field(default_factory=uuid4)

    # Item ID on which this action was performed
    item_id: str

    # User ID who performed this action
    user_id: str

    # the type of action, ex: LIKE, READ, COMMENT etc.
    type: str

    # A numerical value, used by scoring algorithms
    value: float

    # Properties dictionary to hold various metadata such as tags, categories, authors etc.
    props: Optional[Dict]

    # Time when the item was created
    created_at: datetime = Field(default_factory=datetime.now)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
Date: 2022/08/04 12:49
Desc: Yodu.ai  A General purpose Open Source Recommendation Engine
"""
import ast
import re

import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


def get_version_string() -> str:
    """
    Get Yodu.ai version number
    :return: version number
    :rtype: str, e.g. '0.6.24'
    """
    with open("yodu/__init__.py", "rb") as _f:
        version_line = re.search(
            r"__version__\s+=\s+(.*)", _f.read().decode("utf-8")
        ).group(1)
        return str(ast.literal_eval(version_line))


setuptools.setup(
    name="yodu",
    version=get_version_string(),
    author="Shashank Agarwal",
    author_email="shashank@thegeeklabs.com",
    description="Generic Purpose Open Source Recommender System",
    license="Apache 2.0",
    keywords=[
        "recommendations",
        "recommendation",
        "recommenders",
        "recommender",
        "system",
        "engine",
        "machine learning",
        "python",
        "AI",
        "ML",
    ],
    url="https://github.com/thegeeklabs/yodu.ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "pydantic",
        "typing",
        "requests",
        "elasticsearch",
        "python-dotenv",
        "ijson",
    ],
    package_data={"": ["*.py", "*.json", "*.pk", "*.js", "*.zip"]},
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires=">=3.7",
)



from typing import Dict, Optional

from pydantic import BaseModel, Field


class Request(BaseModel):
    # Any fields defined in props will be used to overwrite Provider's default properties
    props: Optional[Dict]

    # user_id of the User for whom items are to be recommended
    user_id: Optional[str]

    # limit on how many items are needed
    limit: Optional[int] = Field(default=10)

    # pagination variable to offset starting point
    offset: Optional[int] = Field(default=0)

    # Internal next_token used by some providers
    next_token: Optional[str]
